from ruxit.api.base_plugin import RemoteBasePlugin
import logging
from ruxit.api.exceptions import ConfigException
from OpenSSL import SSL
from OpenSSL import crypto
from cryptography import x509
from cryptography.hazmat.backends import default_backend
import idna
from socket import socket
from socket import gethostbyname
from base64 import b64encode
import requests
import time
import random
from urllib.parse import quote
import pickle
import tempfile
import os
import sys
from datetime import datetime
from datetime import timedelta
from dataclasses import dataclass
from DTUtils import Logger
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

logger = logging.getLogger(__name__)

domainCheckInterval = 5

ApiURL = "/api/v2/metrics/query?pageSize=1000&resolution=Inf&from=now-1h&metricSelector="

subFolder = "pluginCache"

@dataclass
class Status:
    date: datetime = None
    error: bool = False
    warning: bool = False


class sslCheckPlugin(RemoteBasePlugin):


    def initialize(self, **kwargs):

        # Place all the properties in a dictionary and validate

        self.logger = Logger(self.config["debug"], __name__)   

        self.properties = {}

        self.properties["DomainsToIgnore"] = self.config["DomainsToIgnore"].strip()
        
        if self.config["ClusterAddress"].strip() == "":
            raise ConfigException("Cluster address cannot be empty")
        else:
            self.properties['ClusterAddress'] = self.config["ClusterAddress"].strip()

        if self.properties["ClusterAddress"].rfind("/") == len(self.properties["ClusterAddress"])-1:
            self.properties["ClusterAddress"] = self.properties["ClusterAddress"][:-1]

        self.properties['ExpirationDelta'] = self.config["ExpirationDelta"]


        if self.config["ApiToken"].strip() =="":
            raise ConfigException("Api token cannot be empty")
        else:
            self.properties['ApiToken'] = self.config["ApiToken"].strip()

        self.properties['InboundMetricNames'] = self.config["InboundMetricNames"].strip()

        self.properties['OutboundMetricNames'] = self.config["OutboundMetricNames"].strip()

        self.properties['IgnoreSelfSigned'] = self.config['IgnoreSelfSigned']

            
        self.iterationCount = 0
        
        self.ignoredDomains = [item.strip() for item in self.properties["DomainsToIgnore"].split(',')]  

        self.tempdir = tempfile.gettempdir() + os.path.sep + subFolder + os.path.sep
        self.logger.debug("Cache file location: " + str(self.tempdir))     
        
        self.firstRun = True
        
        self.extraDomains = [item.strip() for item in self.config["CertList"].split(',')] 

        self.proxyName = self.config["ProxyName"].strip()
        
        self.proxyPort = self.config["ProxyPort"]

        self.proxyUser = self.config["ProxyUser"].strip()

        self.proxyPassword = self.config["ProxyPassword"]

        self.proxies = {}
        

        # optional proxy
        self.httpsProxy = self.config.get("ProxySSL", True)

        self.useProxy = self.config.get("UseProxy", False)

        if self.proxyName != "":
            self.logger.info("Setting up proxy")
            if self.proxyUser != "":
                if self.proxyPassword == "":
                    raise ConfigException("please enter a proxy password")
                self.request_proxy = f'{self.proxyUser}:{self.proxyPassword}@{self.proxyName}:{self.proxyPort}/'
            else:
                self.request_proxy = f'{self.proxyName}:{self.proxyPort}/'

            if self.httpsProxy is True:
                self.proxies = { 'https' : "https://" + self.request_proxy, 'http' : "https://" + self.request_proxy} 
            else:
                self.proxies = { 'https' : "http://" + self.request_proxy, 'http' : "http://" + self.request_proxy} 
        else:
            if self.useProxy is True:
                raise ConfigException("No proxy has been configured. Remove the use proxy option or configure a proxy")        
        
        # To ensure that when domains are removed from the list, they are removed from the dictionary too, persist the previous state
        self.mergeDictionary(self.extraDomains)
        


    # store all the extra domains in a file to help detect removal of certs etc...
    def mergeDictionary(self, domainList):
        try:
            # retrieve the previous version if available
            
            confFile = self.tempdir + "prevState.pkl"    
            
            cacheFile = self.loadDictionary()        
            
            oldList = []
            if os.path.exists(confFile):
                with open(confFile, "rb") as out_file:   
                    oldList = pickle.load(out_file) 


            self.logger.debug("old list:" + str(oldList))

            # produce list of items which have been removed.         
            list_dif = [i for i in oldList + domainList if i not in oldList or i not in domainList]
            
            self.logger.debug("diff:" + str(list_dif))
             
            for item in list_dif:
                if item in cacheFile:
                    self.logger.debug("removing:" + str(item))
                    del cacheFile[item]
                       
            for item in domainList:
                # only add if not already in the cache
                if item not in cacheFile and item != "":
                    self.logger.info("Adding:" + str(item))
                    cacheFile[item] = Status()
                    myTime = time.time() + random.randrange(60*10, 60*10+60*60*24)
                    cacheFile[item].date = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(myTime))

            # overwrite with latest values
            with open(confFile, 'wb') as f:
                pickle.dump(domainList, f, pickle.HIGHEST_PROTOCOL)
                
            self.persistDictionary(cacheFile)
            
        except Exception as e:
            self.logger.error("Merge dictionary:" + str(e))
            self.logger.error ('Error on line {}'.format(sys.exc_info()[-1].tb_lineno))
        
        
 
    def query(self, **kwargs):
        try:   
        
            group_name = "SSLCertificates"
            group = self.topology_builder.create_group(group_name, group_name)
            device_name = "SSLPlugin"
            self.device = group.create_device(device_name, device_name)
            
            #self.logger.info(__name__ + ":Topology: group name=%s, device name=%s", group.name, device_name)
            self.iterationCount+=1
            
            # load dictionary from cached file
            self.domainDict = self.loadDictionary()
            
            
            if self.firstRun or self.iterationCount % domainCheckInterval  == 0:
                self.logger.info("Populating domains") 
                self.populateDomains()    
             
            if self.firstRun is True:
                self.firstRun = False  
                
            # write the total of certificates which have been discovered
            self.device.absolute(key='SSLCertCount', value=len(self.domainDict))   
                
            self.checkCertificates(group)     
            
            self.persistDictionary(self.domainDict)
            
        except Exception as e:
            self.logger.info("Query error:" + str(e))
            self.logger.error ('Error on line {}'.format(sys.exc_info()[-1].tb_lineno))

    # load the list of domains that have been discovered so far into memory
    def loadDictionary(self):
        myDict = {}
        try:
            if not os.path.exists(self.tempdir):
                os.makedirs(self.tempdir)
            confFile = self.tempdir + "cache.pkl"
            with open(confFile, "rb") as out_file:
                myDict = pickle.load(out_file)
        except Exception as e:
            self.logger.error("Error loading cache file:" + str(e))
        return myDict

    # Persists the dictionary with the domains to disk as some domains might not appear for a while and not be returned by the API    
    def persistDictionary(self, obj):
        confFile = self.tempdir + "cache.pkl"
        try:
            with open(confFile, 'wb') as f:
                pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        except Exception as e:
            self.logger.error("Error writing cache file:" + str(e)) 
    
    # check that all the properties are not empty.
    def validateProperties(self, config, field):
        value = config[field].strip()
        if value == '':
            raise ConfigException(field.capitalize() + " cannot be empty")
        else:
            self.properties[field] = value
    
    
    # check if domain should be ignored. Works with subdomains.
    # the domains to ignore are passed via the configuration screen. It is a comma separated list of strings to match  
    def ignoreDomain(self, domain):
        if self.ignoredDomains != [""]:
            for items in self.ignoredDomains:
                if domain.find(items)!= -1:
                    return True
        return False
            

        
    # create a dictionary of domains by calling the time series API
    def populateDomains(self):
        
        metrics = [metric.strip() for metric in self.properties["InboundMetricNames"].split(',')]
        for metricName in metrics:
            if metricName != "":
                URL = self.properties['ClusterAddress'] + ApiURL + quote(metricName)
             
                self.callAPI(URL,self.domainDict)
            
        metrics = [metric.strip() for metric in self.properties["OutboundMetricNames"].split(',')]
        for metricName in metrics:
            if metricName != "":        
                URL = self.properties['ClusterAddress'] + ApiURL + metricName
             
                self.callAPI(URL,self.domainDict)
            
        self.logger.info("Size certs:" + str(len(self.domainDict)))  
        
    
    
    # returns true if the name is an IP address as we ignore them because they are unlikely to have certificates
    def isAnIPAddress(self, myIP):
        a = myIP.split('.')
        if len(a) != 4:
            return False
        for x in a:
            if not x.isdigit():
                return False
            i = int(x)
            if i < 0 or i > 255:
                return False
        return True        
     
    # invokes the V2 time series api and adds new values to the dictionary 
    # The key is the domain (could include a port) and the value is when the cert should be checked (once per day)
    # it takes the discovery data and add number number of seconds between 0 and 24 hours later to allow an even distribution of the cert checks.  
    def callAPI(self, URL, domainDict):
        PARAMS = {'Api-Token': self.properties['ApiToken']} 
        try:     
            r = requests.get(url = URL, params = PARAMS, proxies=self.proxies, verify=False) 
            self.logger.debug("Calling API:" + URL)
         
            data = r.json() 
            values = data['result'][0]['data']
            
            if len(values) >= 1000:
                self.device.report_availability_event(title="SSL check",
                                   description=f"More than 1000 splits have been detected for {URL}, please contact your guardian as the data will now be truncated" )        

            
            for item in values:
                dimValue = item['dimensions'][1]
                # some environments return the value twice with a null in the dimension so ignore the null.
                if dimValue is None:
                    continue
                domain = dimValue.split(":")[0]
                
                if self.isAnIPAddress(domain) is False and domain not in domainDict:
                    if not self.ignoreDomain(domain):
                        # randomly distribute the check time on a 24 hours window
                        
                        self.logger.debug("Adding domain from pp:" + str(domain))
                        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        myTime = time.time() + random.randrange(60*10, 60*10+60*60*24)
                        #myTime = time.time() 
                        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        
                        domainDict[domain] = Status()
                        domainDict[domain].date = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(myTime))
                        
                        
        except Exception as e:
            self.logger.error("Error connecting to API:" + URL + " " + str(e))
            self.logger.error ('Error on line {}'.format(sys.exc_info()[-1].tb_lineno))
     
     
    # DNS lookup to see if the domain exists 
    def checkDomainExists(self, domainName):
        try:
            gethostbyname(domainName)
            return True  
        except Exception:
            return False   


    # checks whether or not a certificate is self signed. Inspect the authority, if missing or equal to the subject is should be self signed...
    def isSelfSigned(self, certificate):
        if certificate is None:
            raise Exception ("No Certificate chain detected")
        
        try:
            cert = cert = x509.load_pem_x509_certificate(crypto.dump_certificate( SSL.FILETYPE_PEM,certificate[0].to_cryptography()), default_backend())
            ski = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.SUBJECT_KEY_IDENTIFIER)
            ski2 = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.AUTHORITY_KEY_IDENTIFIER)
            if ski == ski2:
                return True
            else:
                return False
            #self.logger.info("ski:" + str(ski.value.digest) + "   :   " + str(ski2.value.key_identifier))

        # if there is no subject or authority an exception is raised. 
        except Exception as e:
            #self.logger.info ("exception:" + str(e))
            
            return True


    # verify the certificate chain for expiration issues
    # return false is the domain cannot be found and therefore the value is removed.
    # passes custom device group to create a device for each one. This way, alert can be treated separately by Davis.
    def verifyExpiry(self, fullHostname, group):
        
        deconstruct = fullHostname.split(":")
        if len(deconstruct) == 2:
            port = int(deconstruct[1])
        else:
            port = 443             
        hostname = deconstruct[0]  
        if self.checkDomainExists (hostname):
        
            sock = socket()

            self.logger.info("Verifying " + hostname)

            sock_ssl = None
            try:
    
                hostname_idna = idna.encode(hostname)
                myDevice = group.create_device(hostname, hostname)
                
                sock.settimeout(5)
                
                if self.proxyName != ""  and self.useProxy is True:   
                    proxy_name = idna.encode(self.proxyName)
                    sock.connect((proxy_name, self.proxyPort))
                    if self.proxyUser == "":
                        authenticationStr = ""
                    else:
                        tmp = self.proxyUser + ":" + self.proxyPassword
                        authenticationStr = "Proxy-Authorization: Basic " + (b64encode(bytearray(tmp.encode()))).decode() + "\n\r"
                    
                    CONNECT = "CONNECT %s:%s HTTP/1.0\r\n%sConnection: close\r\n\r\n" % (hostname, port, authenticationStr)
                    sock.send(str.encode(CONNECT))
                    buf = sock.recv(8192) 
                else:
                    sock.connect((hostname, port))
                    
                sock.setblocking(1)
                
                ctx = SSL.Context(SSL.SSLv23_METHOD) # most compatible
                ctx.check_hostname = False
                ctx.verify_mode = SSL.VERIFY_NONE


                sock_ssl = SSL.Connection(ctx, sock)
                sock_ssl.set_connect_state()
                sock_ssl.set_tlsext_host_name(hostname_idna)
                sock_ssl.do_handshake()

                
                chain = sock_ssl.get_peer_cert_chain()
                selfSigned = self.isSelfSigned(chain)

                if (selfSigned  is True and self.properties['IgnoreSelfSigned'] is False) or (selfSigned is False):
                    now = datetime.now()

                    for mycert in chain:
                        if mycert is not None:
                            commonName = mycert.get_subject().commonName
                            #issuer = mycert.get_issuer()
                            #self.logger.info("Issuer:" + str(issuer))
                            fullDate = mycert.get_notAfter().decode("utf-8")[0:12]
                            datetime_object = datetime.strptime(fullDate, '%Y%m%d%H%M')
                            timeDelta = (datetime_object- now).days
                            #self.logger.info (timeDelta)
                            if timeDelta < 0:
                                self.logger.info(f"CN: {commonName} expired for {fullHostname} : {fullDate}")
                                self.domainDict[fullHostname].error = True
                                myDevice.report_availability_event(title="SSL check expired",
                                       description=f"The certificate for {fullHostname} has expired" ,
                                       properties={"CN" : commonName ,  "Expiration date": fullDate})                           
                                break
                            elif timeDelta <= int(self.properties['ExpirationDelta']):
                                self.logger.info(f"CN: {commonName} expiring soon for {fullHostname} : {fullDate}")
                                self.domainDict[fullHostname].warning = True
                                myDevice.report_resource_contention_event(title="SSL check expiration warning",
                                       description=f"The certificate for {fullHostname} will expire soon" ,
                                       properties={"CN" : commonName ,  "Expiration date": fullDate})                          
                                break
                            else:
                                self.domainDict[fullHostname].warning = False
                                self.domainDict[fullHostname].error = False

                                #self.logger.info(f"Certificate is good {fullHostname} : {commonName}")
                        else:
                            self.logger.info("Certificate with no expiry for " + hostname)
                else:
                    self.logger.info("Cert ignored as self signed " + fullHostname)
                    
            except Exception as e:
                self.logger.info(__name__+ ":Couldn't retrieve certificate: " + fullHostname + " " + str(e))
                self.logger.error ('Error on line {}'.format(sys.exc_info()[-1].tb_lineno))
            finally:
                if sock_ssl is not None:
                    sock_ssl.close()
                    sock.close()
            return True
        else:
            self.logger.info("Invalid domain " + hostname )
            return False
        
    # iterates through the dictionary and check the last validated date for each entry to see if a check is due  (once per day)
    def checkCertificates(self, group):

        maxChecks = 8
        localCounter = 0
        
        # put a check to avoid checking more than 10 certs (5 sec timeout * 10)
        for certs in list (self.domainDict.keys()):
            checkDate = datetime.strptime(self.domainDict[certs].date, '%Y-%m-%d %H:%M:%S')
            
            #self.logger.info("Cert " + certs + " date " + str(checkDate))
 
 
            #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            now = datetime.now()
            #now = datetime.now() + timedelta(days=1)
            #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

            if (now-checkDate).total_seconds() > 60:
                #only check maxChecks certs per iteration maximum
                if localCounter >= maxChecks:
                    self.logger.info("Reached maximum number of checks per minute")
                    break
                
                # due for a daily check
                self.logger.info(f" {certs} is due for a check")
                if self.verifyExpiry(certs, group):
                    # re-check in 24 hours, check the difference in days between the 2 dates in case, the plugin hasn't run during several days.
                    timeDiff = abs((checkDate - now).days)

                    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    myTime = checkDate + timedelta(days=timeDiff)
                    #myTime = checkDate + timedelta(hours=1)
                    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    
                    #self.logger.info("new date " + myTime.strftime('%Y-%m-%d %H:%M:%S') )
                    self.domainDict[certs].date = myTime.strftime('%Y-%m-%d %H:%M:%S') 
                else:
                    #DNS no longer resolves so remove from list as probably obsolete
                    del (self.domainDict[certs])    
                    
                localCounter += 1

        errorCount = 0
        warningCount = 0
        
        # sum up all the bad certs.
        for items in self.domainDict:
            if self.domainDict[items].error is True:
                errorCount += 1
            if self.domainDict[items].warning is True:
                warningCount += 1 
        

        self.device.absolute(key='SSLCertErrorCount', value=errorCount)
        self.device.absolute(key='SSLCertWarningCount', value=warningCount)  
        
